# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

from math import pi
from libc.math cimport sqrt, sin, cos, acos
import re


def get_color(color):
    b = color & 0xFF
    g = (color & 0xFF00) >> 8
    r = (color & 0xFF0000) >> 16
    return r, g, b

def make_color(r, g, b):
    return b | (g << 8) | (r << 16)

EPSILON = 0.0000001

def coordinates(data):
    if data is None:
        raise ValueError("data is None")
    if len(data) != 2:
        raise ValueError("invalid data length")
    x = (ord(data[0].lower()) - ord('a')) * 64
    y = (int(data[1]) - 1) * 64
    if x < 0 or x >= 512 or y < 0 or y >= 512:
        raise ValueError("coordinates are out-of-bounds")
    return x, y

def to_coordinates(x, y):
    return '{}{}'.format(chr(ord('A') + int(x // 64)), int(y // 64) + 1)

def prettify_timespan(total, get_seconds = False):
    total = int(total)
    if total < 60: get_seconds = True
    days = total // (1440 * 60)
    total -= days * 1440 * 60
    hours = total // (60 * 60)
    total -= hours * 60 * 60
    minutes = total // 60
    seconds = total - minutes * 60 if get_seconds else 0
    if days == hours == minutes == seconds == 0:
        return 'less than a %s' % 'second' if get_seconds else 'minute'
    days_s = '%s day' % days if days > 0 else None
    hours_s = '%s hour' % hours if hours > 0 else None
    minutes_s = '%s minute' % minutes if minutes > 0 else None
    seconds_s = '%s second' % seconds if seconds > 0 else None
    if days > 1: days_s += 's'
    if hours > 1: hours_s += 's'
    if minutes > 1: minutes_s += 's'
    if seconds > 1: seconds_s += 's'
    text = ', '.join([s for s in (days_s, hours_s, minutes_s, seconds_s) if s])
    return text

import zlib
def crc32(data):
    return zlib.crc32(data) & 0xffffffff

# Ace of Spades uses the CP437 character set for chat and Windows-1252 for
# player list

# OpenSpades (and others) use 0xFF as a magic byte indicating the string is UTF-8 encoded

def encode(value):
    if value is not None:
        try:
            return value.encode('cp437', 'strict')
        except UnicodeError:
            return b'\xFF' + value.encode('utf-8', 'replace')

def decode(value):
    if value is not None:
        if value[0] == 0xFF:
            try:
                return value[1:].decode('utf-8', 'strict')
            except UnicodeError: # fallback...
                pass
        return value.decode('cp437', 'replace')

# Printing untrusted ASCII control codes to the console can have a number of
# annoying to dangerous side effects depending on the terminal emulator used

ascii_control_code_regexp = re.compile(r'[\x00-\x1F]')

def _replace_hex(match):
    return r"\x{:x}".format(ord(match.group()))

def escape_control_codes(untrusted_str):
    """escape all ascii control codes in a string note: this still leaves
    things like special unicode characters in place"""

    # Escapes common symbols
    # untrusted_str = untrusted_str.translate({13: '\\r', 10: '\\n', 9: '\\t'})
    # py2 doesn't support the prettier implementation
    untrusted_str = untrusted_str.replace('\r', '\\r').replace('\t', '\\t').replace('\n', '\\n')

    # Escape all characters under 0x20 with their hex notation
    return ascii_control_code_regexp.sub(_replace_hex, untrusted_str)

cdef class Vertex3:
    # NOTE: for the most part this behaves as a 2d vector, with z being tacked on
    # so it's useful for orientation math
    def __init__(self, *arg, is_ref = False):
        cdef float x, y, z
        if not is_ref:
            if arg:
                x, y, z = arg
            else:
                x = y = z = 0.0
            self.value = new Vector(x, y, z)
        self.is_ref = is_ref

    def __dealloc__(self):
        if not self.is_ref:
            del self.value
            self.value = NULL

    def copy(self):
        return create_vertex3(self.value.x, self.value.y, self.value.z)

    def get(self):
        # temp vars are needed to disable cython tuple optimization
        tmp_x = self.value.x
        tmp_y = self.value.y
        tmp_z = self.value.z
        return tmp_x, tmp_y, tmp_z

    def set(self, float x, float y, float z):
        self.value.x = x
        self.value.y = y
        self.value.z = z

    def set_vector(self, Vertex3 vector):
        self.value.x = vector.value.x
        self.value.y = vector.value.y
        self.value.z = vector.value.z

    def zero(self):
        self.value.x = self.value.y = self.value.z = 0.0

    def __add__(self, Vertex3 A):
        cdef Vector * a = (<Vertex3>self).value
        cdef Vector * b = A.value
        return create_vertex3(a.x + b.x, a.y + b.y, a.z + b.z)

    def __sub__(self, Vertex3 A):
        cdef Vector * a = (<Vertex3>self).value
        cdef Vector * b = A.value
        return create_vertex3(a.x - b.x, a.y - b.y, a.z - b.z)

    def __mul__(self, float k):
        cdef Vector * a = (<Vertex3>self).value
        return create_vertex3(a.x * k, a.y * k, a.z * k)

    def __truediv__(self, float k):
        cdef Vector * a = (<Vertex3>self).value
        return create_vertex3(a.x / k, a.y / k, a.z / k)

    def __iadd__(self, Vertex3 A):
        cdef Vector * a = self.value
        cdef Vector * b = A.value
        a.x += b.x
        a.y += b.y
        a.z += b.z
        return self

    def __isub__(self, Vertex3 A):
        cdef Vector * a = self.value
        cdef Vector * b = A.value
        a.x -= b.x
        a.y -= b.y
        a.z -= b.z
        return self

    def __imul__(self, double k):
        cdef Vector * a = self.value
        a.x *= k
        a.y *= k
        a.z *= k
        return self

    def __idiv__(self, double k):
        cdef Vector * a = self.value
        a.x /= k
        a.y /= k
        a.z /= k
        return self

    def translate(self, double x, double y, double z):
        cdef Vector * a = self.value
        a.x += x
        a.y += y
        a.z += z
        return self

    def cross(self, Vertex3 A):
        cdef Vector * a = self.value
        cdef Vector * b = A.value
        return create_vertex3(
            a.y * b.z - a.z * b.y,
            a.z * b.x - a.x * b.z,
            a.x * b.y - a.y * b.x)

    def dot(self, Vertex3 A):
        cdef Vector * a = self.value
        cdef Vector * b = A.value
        return a.x * b.x + a.y * b.y

    def perp_dot(self, Vertex3 A):
        cdef Vector * a = self.value
        cdef Vector * b = A.value
        return a.x * b.y - a.y * b.x

    def rotate(self, Vertex3 A):
        cdef Vector * a = self.value
        cdef Vector * b = A.value
        a.x, a.y = a.x * b.x - a.y * b.y, a.x * b.y + a.y * b.x
        return self

    def unrotate(self, Vertex3 A):
        cdef Vector * a = self.value
        cdef Vector * b = A.value
        a.x, a.y = a.x * b.x + a.y * b.y, a.y * b.x - a.x * b.y
        return self

    def length(self):
        cdef Vector * a = self.value
        return sqrt(a.x * a.x + a.y * a.y + a.z * a.z)

    def distance(self, Vertex3 other):
        """calculate euclidean (straight line) distance between two `Vertex3` as
        points in 3d space. Equivalent to ``(a - b).length()``."""
        cdef Vector * a = self.value
        cdef Vector * b = other.value
        x = a.x - b.x
        y = a.y - b.y
        z = a.z - b.z
        return sqrt(x**2 + y**2 + z**2)

    def length_sqr(self):
        """calculate the square length of the vertex3. This is a bit faster than
        getting the length, as it avoids the expensive `sqrt` call."""
        cdef Vector * a = self.value
        return a.x * a.x + a.y * a.y + a.z * a.z

    def is_zero(self):
        return self.length_sqr() < EPSILON

    def normal(self):
        cdef float k = self.length()
        cdef Vector * a = self.value
        return k and create_vertex3(a.x / k, a.y / k, a.z / k) or Vertex3()

    def normalize(self):
        cdef float k = self.length()
        cdef Vector * a = self.value
        if k:
            a.x /= k
            a.y /= k
            a.z /= k
        else:
            a.x = 0
            a.y = 0
            a.z = 0
        return k

    cpdef Quaternion get_rotation_to(self, Vertex3 A):
        q = Quaternion()
        v0 = self.normal()
        v1 = A.normal()
        d = v0.dot(v1)
        if d >= 1.0:
            return Quaternion()
        if d < EPSILON - 1.0:
            axis = Vertex3(1.0, 0.0, 0.0).cross(self)
            if axis.is_zero():
                axis = Vertex3(0.0, 1.0, 0.0).cross(self)
            axis.normalise()
            q.set_angle_axis(pi, axis)
        else:
            k = sqrt((1.0 + d) * 2.0)
            inv_k = 1.0 / k
            c = v0.cross(v1)
            q.x = c.x * inv_k
            q.y = c.y * inv_k
            q.z = c.z * inv_k
            q.w = k * 0.5
            q.normalize()
        return q

    def __neg__(self):
        cdef Vector * a = self.value
        return create_vertex3(-a.x, -a.y, -a.z)

    def __pos__(self):
        cdef Vector * a = self.value
        return create_vertex3(+a.x, +a.y, +a.z)

    def __repr__(self):
        cdef Vector * a = self.value
        # "5 decimal places ought to be enough for anyone (tm)"
        return "Vertex3({:.5f}, {:.5f}, {:.5f})".format(a.x, a.y, a.z)

    # properties for python wrapper
    property x:
        def __get__(self):
            return self.value.x
        def __set__(self, value):
            self.value.x = value

    property y:
        def __get__(self):
            return self.value.y
        def __set__(self, value):
            self.value.y = value

    property z:
        def __get__(self):
            return self.value.z
        def __set__(self, value):
            self.value.z = value

cdef class Quaternion:
    def __init__(self, *arg):
        self.w = 1.0
        if arg:
            self.set(*arg)

    def copy(self):
        return Quaternion(self.w, self.x, self.y, self.z)

    def get(self):
        return self.w, self.x, self.y, self.z

    def set(self, double w, double x, double y, double z):
        self.w = w
        self.x = x
        self.y = y
        self.z = z

    cpdef Quaternion set_angle_axis(self, double radians, Vertex3 axis):
        # axis must be normalized
        half_angle = radians * 0.5
        sha = sin(half_angle)
        self.w = cos(half_angle)
        self.x = axis.x * sha
        self.y = axis.y * sha
        self.z = axis.z * sha
        return self

    cpdef Quaternion slerp(self, Quaternion q, double t):
        if t <= 0.0: return self.copy()
        if t >= 1.0: return q.copy()

        cos_omega = self.x * q.x + self.y * q.y + self.z * q.z + self.w * q.w
        k0, k1 = 1.0 - t, t
        if cos_omega < 0.0:
            q.w, q.x, q.y, q.z = -q.w, -q.x, -q.y, -q.z
            cos_omega = -cos_omega
        if cos_omega <= 0.9999:
            omega = acos(cos_omega)
            sin_omega = sin(omega)
            k0 = sin(k0 * omega) / sin_omega
            k1 = sin(k1 * omega) / sin_omega

        return Quaternion(
            k0 * self.w + k1 * q.w,
            k0 * self.x + k1 * q.x,
            k0 * self.y + k1 * q.y,
            k0 * self.z + k1 * q.z)

    cpdef Quaternion nlerp(self, Quaternion q, double t):
        return (self.multiply_scalar(1.0 - t) + q.multiply_scalar(t)).normalize()

    cpdef Vertex3 transform_vector(self, Vertex3 v):
        tx = self.w * v.x + self.y * v.z - self.z * v.y
        ty = self.w * v.y - self.x * v.z + self.z * v.x
        tz = self.w * v.z + self.x * v.y - self.y * v.x
        tw = -(self.x * v.x + self.y * v.y + self.z * v.z)

        return Vertex3(
            -tw * self.x + tx * self.w - ty * self.z + tz * self.y,
            -tw * self.y + ty * self.w - tz * self.x + tx * self.z,
            -tw * self.z + tz * self.w - tx * self.y + ty * self.x)

    cpdef Vertex3 inverse_transform_vector(self, Vertex3 v):
        tx = self.w * v.x - self.y * v.z + self.z * v.y;
        ty = self.w * v.y + self.x * v.z - self.z * v.x;
        tz = self.w * v.z - self.x * v.y + self.y * v.x;
        tw = self.x * v.x + self.y * v.y + self.z * v.z;

        return Vertex3(
            tw * self.x + tx * self.w + ty * self.z - tz * self.y,
            tw * self.y + ty * self.w + tz * self.x - tx * self.z,
            tw * self.z + tz * self.w + tx * self.y - ty * self.x)

    def normalize(self):
        k = self.w * self.w + self.x * self.x + self.y * self.y + self.z * self.z
        if abs(k - 1.0) > EPSILON:
            k = 1.0 / sqrt(k)
            self.w *= k
            self.x *= k
            self.y *= k
            self.z *= k
        return self

    def multiply_scalar(self, double k):
        return Quaternion(self.w * k, self.x * k, self.y * k, self.z * k)

    def get_matrix(self):
        x2, y2, z2 = self.x * self.x, self.y * self.y, self.z * self.z
        xy = self.x * self.y
        xz = self.x * self.z
        yz = self.y * self.z
        wx = self.w * self.x
        wy = self.w * self.y
        wz = self.w * self.z

        # column-major
        return (1.0 - 2.0 * (y2 + z2), 2.0 * (xy - wz), 2.0 * (xz + wy), 0.0,
            2.0 * (xy + wz),1.0 - 2.0 * (x2 + z2), 2.0 * (yz - wx), 0.0,
            2.0 * (xz - wy), 2.0 * (yz + wx), 1.0 - 2.0 * (x2 + y2), 0.0,
            0.0, 0.0, 0.0, 1.0)

    def __add__(self, Quaternion A):
        return Quaternion(self.w + A.w, self.x + A.x, self.y + A.y, self.z + A.z)

    def __mul__(self, Quaternion A):
        return Quaternion(
            self.w * A.w - self.x * A.x - self.y * A.y - self.z * A.z,
            self.w * A.x + self.x * A.w + self.y * A.z - self.z * A.y,
            self.w * A.y + self.y * A.w + self.z * A.x - self.x * A.z,
            self.w * A.z + self.z * A.w + self.x * A.y - self.y * A.x)

    def __neg__(self):
        return Quaternion(self.w, -self.x, -self.y, -self.z)

    def __pos__(self):
        return Quaternion(self.w, +self.x, +self.y, +self.z)

    def __str__(self):
        return "(%s %s %s %s)" % (self.w, self.x, self.y, self.z)



import math
from typing import Callable, Dict
from twisted.internet import reactor
from pyspades.constants import (RIFLE_WEAPON, SMG_WEAPON, SHOTGUN_WEAPON,
                                HEAD, TORSO, ARMS, LEGS, CLIP_TOLERANCE)


class BaseWeapon:
    shoot = False
    reloading = False
    id = None  # type: int
    shoot_time = None
    next_shot = 0
    start = None

    # Weapon parameters
    # Total number of rounds
    stock = None  # type: int
    # Number of rounds that fit in the weapon at once
    ammo = None  # type: int
    # If the weapon should be reloaded one round at a time like the shotgun
    slow_reload = False
    # Time between shots
    delay = 0.0
    # Time between reloading and being able to shoot again
    reload_time = 0.0
    # Dict of damages
    damage = None  # type: Dict[int, int]

    def __init__(self, reload_callback: Callable) -> None:
        self.reload_callback = reload_callback
        self.reset()

    def restock(self) -> None:
        self.current_stock = self.stock

    def reset(self) -> None:
        self.shoot = False
        if self.reloading:
            self.reload_call.cancel()
            self.reloading = False
        self.current_ammo = self.ammo
        self.current_stock = self.stock

    def set_shoot(self, value: bool) -> None:
        if value == self.shoot:
            return
        current_time = reactor.seconds()
        if value:
            self.start = current_time
            if self.current_ammo <= 0:
                return
            elif self.reloading and not self.slow_reload:
                return
            self.shoot_time = max(current_time, self.next_shot)
            if self.reloading:
                self.reloading = False
                self.reload_call.cancel()
        else:
            ammo = self.current_ammo
            self.current_ammo = self.get_ammo(True)
            self.next_shot = self.shoot_time + self.delay * (
                ammo - self.current_ammo)
        self.shoot = value

    def reload(self) -> None:
        if self.reloading:
            return
        ammo = self.get_ammo()
        if not self.current_stock or ammo >= self.ammo:
            return
        elif self.slow_reload and self.shoot and ammo:
            return
        self.reloading = True
        self.set_shoot(False)
        self.current_ammo = ammo
        self.reload_call = reactor.callLater(self.reload_time, self.on_reload)

    def on_reload(self) -> None:
        self.reloading = False
        if self.slow_reload:
            self.current_ammo += 1
            self.current_stock -= 1
            self.reload_callback()
            self.reload()
        else:
            new_stock = max(0, self.current_stock - (
                self.ammo - self.current_ammo))
            self.current_ammo += self.current_stock - new_stock
            self.current_stock = new_stock
            self.reload_callback()

    def get_ammo(self, no_max: bool = False) -> int:
        if self.shoot:
            dt = reactor.seconds() - self.shoot_time
            ammo = self.current_ammo - max(0, int(
                math.ceil(dt / self.delay)))
        else:
            ammo = self.current_ammo
        if no_max:
            return ammo
        return max(0, ammo)

    def is_empty(self, tolerance=CLIP_TOLERANCE) -> bool:
        return self.get_ammo(True) < -tolerance or not self.shoot

    def get_damage(self, value, position1, position2) -> int:
        return self.damage[value]


class Rifle(BaseWeapon):
    id = RIFLE_WEAPON
    name = 'Rifle'
    delay = 0.5
    ammo = 10
    stock = 50
    reload_time = 2.5
    slow_reload = False

    damage = {
        TORSO: 49,
        HEAD: 100,
        ARMS: 33,
        LEGS: 33
    }


class SMG(BaseWeapon):
    id = SMG_WEAPON
    name = 'SMG'
    delay = 0.11  # actually 0.1, but due to AoS scheduling, it's usually 0.11
    ammo = 30
    stock = 120
    reload_time = 2.5
    slow_reload = False

    damage = {
        TORSO: 29,
        HEAD: 75,
        ARMS: 18,
        LEGS: 18
    }


class Shotgun(BaseWeapon):
    id = SHOTGUN_WEAPON
    name = 'Shotgun'
    delay = 1.0
    ammo = 6
    stock = 48
    reload_time = 0.5
    slow_reload = True

    damage = {
        TORSO: 27,
        HEAD: 37,
        ARMS: 16,
        LEGS: 16
    }


WEAPONS = {
    RIFLE_WEAPON: Rifle,
    SMG_WEAPON: SMG,
    SHOTGUN_WEAPON: Shotgun,
}



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

from pyspades.loaders cimport Loader
from pyspades.bytes cimport ByteReader, ByteWriter

_client_loaders = {}
_server_loaders = {}

def register_packet(loader=None, server=True, client=True, extension=None):
    """register a packet

    >>> @register_packet()
    ... class SomePacket(Loader):
    ...     pass

    Optionally can be used as function too, but this is discouraged unless you
    need to do this (the default pyspades loaders need this, as they are
    defined in C++ code):

    >>> class AnotherPacket(Loader):
    ...     pass
    ...
    >>> register_packet(AnotherPacket, server=False)

    Additionally you can specify if only the server or client can send this
    packet. This is only useful in a few rare cases, so these are set to
    ``True`` by default.

    Arguments:
        server (bool, optional): This packet can be sent by the server. True by
            default.
        client (bool, optional): This packet can be sent by the client. True by
            default
        extension (int, optional): The extension id this packet belongs to, if any

    Raises:
        KeyError: If the packet's ID has already been registered
    """

    # Rationale:
    # In the past, the packet list was static and contained no meta-information
    # about the packet this was fine, because there was really only one
    # version.  However, it is foreseeable from current developments that there
    # will be several played protocol versions and a number of clients could
    # have different feature sets that must be queried at runtime. Hence it is
    # useful to annotate packets with some kind of runtime information.

    # It is not known if full-blown multiple protocol version support will ever
    # be realized, but this is one step along the way. It might be a massive
    # miscalculation and be technical debt for the rest of the projects
    # lifetime, but only time will tell

    def register(cls):
        # Having to handle the two loader dicts separate isn't very nice, but I
        # you can't avoid it whithout making things more complicated somewhere
        # else.
        # This is needed because certain packets (e.g. HitPacket and SetHP) in
        # likely a legendary case of premature optimisation (or because the
        # packet IDs had to be contigous in voxlap?) share packet IDs, but have
        # different meaning and format depending on who sends them
        # Since this library tries to work as both a Server and Client
        # implementation, both are always evaluated

        if client:
            if cls.id in _client_loaders:
                msg = ("Tried to register client packet '{!r}', but a packet "
                       "with that id ({}) already exists.".format(cls, cls.id))
                raise KeyError(msg)
            _client_loaders[cls.id] = cls

        if server:
            if cls.id in _server_loaders:
                msg = ("Tried to register server packet '{!r}', but a packet "
                       "with that id ({}) already exists.".format(cls, cls.id))
                raise KeyError(msg)
            _server_loaders[cls.id] = cls

        return cls

    if loader:
        # Call the registration function directly if we are being called as a
        # function, not as a decorator
        register(loader)

    return register


def load_server_packet(data):
    return load_contained_packet(data, _server_loaders)

def load_client_packet(data):
    return load_contained_packet(data, _client_loaders)

cdef inline Loader load_contained_packet(ByteReader data, dict table):
    type_ = data.readByte(True)
    return table[type_](data)

_packet_handlers = {}

def register_packet_handler(loader):
    def register_handler(function):
        _packet_handlers[loader.id] = function
        return function
    return register_handler

def call_packet_handler(self, loader):
    contained = load_client_packet(ByteReader(loader.data))
    try:
        handler = _packet_handlers[contained.id]
    except KeyError:
        # an invalid ID was sent
        pass
    finally:
        # we call the handler in the finally clause so we don't
        # accidentally ignore KeyErrors the handler raises
        handler(self, contained)



#!/usr/bin/env python3

import sys
import os
from distutils.core import run_setup
# build_ext is subclassed, so we import it with a _ to avoid a collision
from distutils.command.build_ext import build_ext as _build_ext
from setuptools import setup, Extension, find_packages


PKG_NAME = "piqueserver"

extra_args = sys.argv[2:]

with open('README.rst') as f:
    long_description = f.read()

# load version info from the piqueserver module manually
here = os.path.abspath(os.path.dirname(__file__))
version = {}
with open(os.path.join(here, 'piqueserver/version.py')) as f:
    exec(f.read(), version)

ext_names = [
    'pyspades.vxl',
    'pyspades.bytes',
    'pyspades.packet',
    'pyspades.contained',
    'pyspades.common',
    'pyspades.world',
    'pyspades.loaders',
    'pyspades.mapmaker'
]

static = os.environ.get('STDCPP_STATIC') == "1"

if static:
    print("Linking the build statically.")

linetrace = os.environ.get('CYTHON_TRACE') == '1'

# Compile the server with support for
# AddressSanitizer/UndefinedBehaviourSanitizer
USE_ASAN = os.environ.get('USE_ASAN') == '1'
USE_UBSAN = os.environ.get('USE_UBSAN') == '1'

ext_modules = []

for name in ext_names:
    extra = {
        "define_macros": [],
        "extra_link_args": [],
        "extra_compile_args": [],
    }  # type: dict

    if static:
        extra['extra_link_args'].extend(
            ['-static-libstdc++', '-static-libgcc'])

    if USE_ASAN:
        extra["extra_link_args"].append("-lasan")
        extra["extra_compile_args"].append("-fsanitize=address")

    if USE_UBSAN:
        extra["extra_link_args"].append("-lubsan")
        extra["extra_compile_args"].append("-fsanitize=undefined")

    if name in ['pyspades.vxl', 'pyspades.world', 'pyspades.mapmaker']:
        extra["extra_compile_args"].append('-std=c++11')

    if linetrace:
        extra['define_macros'].append(('CYTHON_TRACE', '1'))

    ext_modules.append(Extension(name, ['./%s.pyx' % name.replace('.', '/')],
                                 language='c++', include_dirs=['./pyspades'],
                                 **extra))


class build_ext(_build_ext):

    def run(self):

        from Cython.Build import cythonize

        if USE_ASAN:
            from Cython.Compiler import Options
            # make asan/valgrind's memory leak results better
            Options.generate_cleanup_code = True

        compiler_directives = {'language_level': 3, 'embedsignature': True}
        if linetrace:
            compiler_directives['linetrace'] = True

        self.extensions = cythonize(self.extensions, compiler_directives=compiler_directives)

        _build_ext.run(self)

        run_setup(os.path.join(os.getcwd(), "setup.py"),
                  ['build_py'] + extra_args)


setup(
    name=PKG_NAME,
    packages=find_packages(exclude=("tests", "tests.*")),
    version=version['__version__'],
    description='Open-Source server implementation for Ace of Spades ',
    author=('Originally MatPow2 and PySnip contributors,'
            'now, StackOverflow and piqueserver authors'),
    author_email='nate.shoffner@gmail.com',
    maintainer='noway421',
    maintainer_email='noway@2ch.hk',
    license='GNU General Public License v3',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/piqueserver/piqueserver",
    keywords=['ace of spades', 'aos', 'server',
              'pyspades', 'pysnip', 'piqueserver'],
    python_requires=">=3.5.3",
    classifiers=[
        'Intended Audience :: System Administrators',
        'Development Status :: 5 - Production/Stable',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 3',
        'Framework :: Twisted',
        'Topic :: Games/Entertainment',
        'Topic :: Games/Entertainment :: First Person Shooters',
    ],
    platforms="Darwin, Unix, Win32",

    setup_requires=['Cython>=0.27,<1'],
    install_requires=[
        'pypiwin32;platform_system=="Windows"',
        'Cython>=0.27,<1',
        'Twisted[tls]',
        'Jinja2>=2,<3',
        'Pillow>=5.1.0,<7',
        'aiohttp>=3.3.0,<3.7.0',
        'pyenet',
        'toml',
        'packaging>=19.0'
    ],
    extras_require={
        'from': ['geoip2>=2.9,<3.0'],
        'ssh': ['Twisted[tls,conch]'],
    },
    entry_points={
        'console_scripts': [
            '%s=%s.run:main' % (PKG_NAME, PKG_NAME)
        ],
    },
    package_data={"%s.web" % PKG_NAME: ["templates/status.html"]},
    include_package_data=True,

    ext_modules=ext_modules,
    cmdclass={'build_ext': build_ext},
)



"""
Gives Ace of Spades a daycycle (using the fog).

Commands
^^^^^^^^

* ``/dayspeed <speed>`` sets day speed *admin only*
* ``/dayspeed`` gets day speed *admin only*
* ``/daytime <time>`` sets day time *admin only*
* ``/daytime`` gets day time

.. codeauthor:: hompy
"""

from math import modf

from twisted.internet.task import LoopingCall
from piqueserver.commands import command, admin
from pyspades.color import wrap, interpolate_hsb, interpolate_rgb, hsb_to_rgb, rgb_distance

S_NO_RIGHTS = 'No administrator rights!'
S_TIME_OF_DAY = 'Time of day: {hours:02d}:{minutes:02d}'
S_SPEED = 'Day cycle speed is {multiplier}'
S_SPEED_SET = 'Day cycle speed changed to {multiplier}'
S_STOPPED = 'Day cycle stopped'


@command('dayspeed', admin_only=True)
def day_speed(connection, value=None):
    if value is None:
        return S_SPEED.format(multiplier=connection.protocol.time_multiplier)
    value = float(value)
    protocol = connection.protocol
    protocol.time_multiplier = value
    if value == 0.0:
        if protocol.daycycle_loop.running:
            protocol.daycycle_loop.stop()
        return S_STOPPED
    else:
        if not protocol.daycycle_loop.running:
            protocol.daycycle_loop.start(protocol.day_update_frequency)
        return S_SPEED_SET.format(multiplier=value)


@command('daytime')
def day_time(connection, value=None):
    if value is not None:
        if not connection.admin:
            return S_NO_RIGHTS
        time = float(value)
        if time < 0.0:
            raise ValueError("Time cannot be < 0")
        connection.protocol.current_time = time
        connection.protocol.update_day_color()
    f, i = modf(connection.protocol.current_time)
    return S_TIME_OF_DAY.format(hours=int(i), minutes=int(f * 60))


def apply_script(protocol, connection, config):
    class DayCycleProtocol(protocol):
        current_color = None
        current_time = None
        daycycle_loop = None
        day_duration = None
        day_update_frequency = None
        time_multiplier = None

        def __init__(self, *arg, **kw):
            protocol.__init__(self, *arg, **kw)
            self.daycycle_loop = LoopingCall(self.update_day_color)
            self.reset_daycycle()

        def reset_daycycle(self):
            if not self.daycycle_loop:
                return
            self.current_color = None
            self.current_time = 7.00
            self.day_duration = 24 * 60 * 60.00
            self.day_update_frequency = 0.1
            self.time_multiplier = 48.0
            self.day_colors = [
                (0.00, (0.05, 0.05, 0.05), False),
                (4.00, (0.05, 0.77, 0.05), False),
                (5.00, (0.0694, 0.77, 0.78), True),
                (5.30, (0.0361, 0.25, 0.95), False),
                (6.00, (0.56, 0.18, 0.94), False),
                (9.00, (0.5527, 0.24, 0.94), False),
                (12.00, (0.5527, 0.41, 0.95), False),
                (19.50, (0.56, 0.28, 0.96), False),
                (20.00, (0.15, 0.33, 0.87), False),
                (20.25, (0.11, 0.49, 0.94), False),
                (20.50, (0.1056, 0.69, 1.00), False),
                (22.50, (0.1, 0.69, 0.1), True),
                (23.00, (0.05, 0.05, 0.05), False)]
            self.time_step = 24.00 / (self.day_duration /
                                      self.day_update_frequency)
            self.target_color_index = 0
            self.next_color()
            if not self.daycycle_loop.running:
                self.daycycle_loop.start(self.day_update_frequency)

        def update_day_color(self):
            if self.current_time >= 24.00:
                self.current_time = wrap(0.00, 24.00, self.current_time)
            while (self.current_time < self.start_time or
                   self.current_time >= self.target_time):
                self.next_color()
                self.target_time = self.target_time or 24.00
            t = ((self.current_time - self.start_time) /
                 (self.target_time - self.start_time))
            if self.hsv_transition:
                new_color = interpolate_hsb(self.start_color,
                                            self.target_color, t)
                new_color = hsb_to_rgb(*new_color)
            else:
                new_color = interpolate_rgb(self.start_color,
                                            self.target_color, t)
            if (self.current_color is None or
                    rgb_distance(self.current_color, new_color) > 3):
                self.current_color = new_color
                self.set_fog_color(self.current_color)
            self.current_time += self.time_step * self.time_multiplier

        def next_color(self):
            self.start_time, self.start_color, _ = (
                self.day_colors[self.target_color_index])
            self.target_color_index = ((self.target_color_index + 1) %
                                       len(self.day_colors))
            self.target_time, self.target_color, self.hsv_transition = (
                self.day_colors[self.target_color_index])
            if not self.hsv_transition:
                self.start_color = hsb_to_rgb(*self.start_color)
                self.target_color = hsb_to_rgb(*self.target_color)

        def on_map_change(self, map_):
            self.reset_daycycle()
            protocol.on_map_change(self, map_)

    return DayCycleProtocol, connection



"""
Zones of control: Dropped intel and tents exert influence
over nearby area, restricting player ability to destroy.

Options
^^^^^^^

.. code-block:: guess

   [zoc]
   radius = 32
   attack_distance = 64
   block_undo = 10
   block_cost = 5
   points_per_tick = 1
   point_cap = 30
   grenade_cost = 30

.. codeauthor: ?
"""

from collections import deque
from twisted.internet.task import LoopingCall
from pyspades.constants import GRENADE_DESTROY, SPADE_DESTROY, CTF_MODE, TC_MODE
from piqueserver.config import config

BK_FREE, BK_FRIENDLY, BK_ENEMY_FAR, BK_ENEMY_NEAR, BK_UNDO = range(5)

zoc_config = config.section("zoc")
zoc_radius = zoc_config.option("radius", 32)
zoc_attack_distance = zoc_config.option("attack_distance", 64)
zoc_block_undo = zoc_config.option("block_undo", 10)

zoc_block_cost = zoc_config.option("block_cost", 5)
zoc_points_per_tick = zoc_config.option("points_per_tick", 1)
zoc_point_cap = zoc_config.option("point_cap", 6 * zoc_block_cost.get())
zoc_grenade_cost = zoc_config.option("grenade_cost", zoc_point_cap.get())


def apply_script(protocol, connection, config):

    class ZOCConnection(connection):
        def on_connect(self):
            self.block_undo = deque()
            self.zoc_destruction_points = 0
            return connection.on_connect(self)

        def on_spawn(self, pos):
            self.zoc_destruction_points = 0
            return connection.on_spawn(self, pos)

        def on_block_destroy(self, x, y, z, mode):
            """Restrict destruction of blocks that are on your own team and
            in a zone of control, or blocks that are on the other team and
            are too distant."""
            if not self.god:
                zoc = self.zoc_type(x, y, z)
                if zoc == BK_ENEMY_FAR:
                    self.send_chat(
                        "You're too far away to attack this area!")
                    return False
                elif zoc == BK_FRIENDLY:
                    cost = zoc_block_cost.get()
                    if mode == SPADE_DESTROY:
                        cost *= 3
                    elif mode == GRENADE_DESTROY:
                        cost = zoc_grenade_cost.get()
                    if self.zoc_destruction_points < cost:
                        self.send_chat("Stop destroying your territory! " +
                                       "Go fight the enemy!")
                        return False
                    else:
                        self.zoc_destruction_points -= cost
            return connection.on_block_destroy(self, x, y, z, mode)

        def own_block(self, x, y, z):
            for position in self.block_undo:
                if position[0] == x and position[1] == y and position[2] == z:
                    return True
            return False

        def on_block_build_attempt(self, x, y, z):
            zoc = self.zoc_type(x, y, z)
            if zoc == BK_ENEMY_NEAR:
                self.send_chat("You can't build in enemy territory!")
                return False
            else:
                return connection.on_block_build_attempt(self, x, y, z)

        def on_block_build(self, x, y, z):
            self.block_undo.append((x, y, z))
            if len(self.block_undo) > zoc_block_undo.get():
                self.block_undo.popleft()
            return connection.on_block_build(self, x, y, z)

        def zoc_type(self, x, y, z):
            for zoc in self.protocol.zone_cache:
                if (zoc['left'] <= x and zoc['right'] >= x and
                        zoc['top'] <= y and zoc['bottom'] >= y):
                    if zoc['team'] is self.team:
                        return BK_UNDO if self.own_block(
                            x, y, z) else BK_FRIENDLY
                    else:
                        p_x, p_y, _ = self.world_object.position.get()
                        dist_sq = (p_x - x) * (p_x - x) +\
                                  (p_y - y) * (p_y - y)
                        if zoc_attack_distance.get()**2 < dist_sq:
                            return BK_ENEMY_FAR
                        else:
                            return BK_ENEMY_NEAR
            return BK_FREE

    class ZOCProtocol(protocol):

        zone_cache = None


        def __init__(self, *arg, **kw):
            # we update the zones with a slow loop
            # (simpler than tracking every event that could update zones)
            protocol.__init__(self, *arg, **kw)
            self.zoc_loop = LoopingCall(self.zoc_tick)
            if not self.zoc_loop.running:
                self.zoc_loop.start(5.0)

        def _build_zoc(self, x, y, team):
            return {'team': team,
                    'left': x - zoc_radius.get(),
                    'right': x + zoc_radius.get(),
                    'top': y - zoc_radius.get(),
                    'bottom': y + zoc_radius.get()}

        def zoc_tick(self):
            self.cache_zones_of_control()
            for player in list(self.players.values()):
                player.zoc_destruction_points += zoc_points_per_tick.get()
                if player.zoc_destruction_points > zoc_point_cap.get():
                    player.zoc_destruction_points = zoc_point_cap.get()

        def cache_zones_of_control(self):
            zones = []
            if self.game_mode == CTF_MODE:
                for flag in [self.green_team.flag, self.blue_team.flag]:
                    if flag.player is None:
                        zones.append(self._build_zoc(
                            flag.x, flag.y, flag.team))
                for base in [self.green_team.base, self.blue_team.base]:
                    zones.append(self._build_zoc(base.x, base.y, base.team))
            elif self.game_mode == TC_MODE:
                for flag in self.entities:
                    zones.append(self._build_zoc(flag.x, flag.y, flag.team))
            self.zone_cache = zones

        def on_update_entity(self, entity):
            desired_z = self.map.get_z(entity.x, entity.y)
            if int(entity.z) != int(desired_z):
                entity.z = desired_z
                return True
            return protocol.on_update_entity(self, entity)

    return ZOCProtocol, ZOCConnection



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

"""
A few useful types used around the place.

IDPool is used to distribute the IDs given out by the Server

AttributeSet is used for testing if various settings are active

MultikeyDict is used to make player names accessible by both id and name
"""

import itertools
from collections import namedtuple


class IDPool:
    """
    Manage pool of IDs

    >>> p = IDPool(start=10)
    >>> p.pop()
    10
    >>> p.pop()
    11
    >>> p.pop()
    12
    >>> p.put_back(11)
    >>> p.pop()
    11
    """

    def __init__(self, start=0):
        self.free_ids = []
        self.new_ids = itertools.count(start)

    def pop(self):
        if self.free_ids:
            return self.free_ids.pop()
        else:
            return next(self.new_ids)

    def put_back(self, id):
        self.free_ids.append(id)


class AttributeSet(set):
    """
    set with attribute access, i.e.

    >>> foo = AttributeSet(("eggs", ))
    >>> foo.eggs
    True
    >>> foo.spam
    False

    Also supports adding and removing elements

    >>> foo.bar = True
    >>> 'bar' in foo
    True
    >>> foo.bar = False
    >>> 'bar' in foo
    False

    This works as a quick shorthand for membership testing.
    """

    def __getattr__(self, name):
        return name in self

    def __setattr__(self, name, value):
        if value:
            self.add(name)
        else:
            self.discard(name)



from math import floor
import colorsys
from typing import Tuple

ColorTuple = Tuple[float, float, float]


def wrap(minimum: float, maximum: float, value: float) -> float:
    return value - floor(
        (value - minimum) / (maximum - minimum)) * (maximum - minimum)


def hsb_to_rgb(hue: float, sat: float, bri: float) -> ColorTuple:
    r, g, b = colorsys.hsv_to_rgb(hue, sat, bri)
    return int(r * 255), int(g * 255), int(b * 255)


def interpolate_rgb(xyz1, xyz2, t):
    (r1, g1, b1) = xyz1
    (r2, g2, b2) = xyz2
    return (int(r1 + (r2 - r1) * t),
            int(g1 + (g2 - g1) * t),
            int(b1 + (b2 - b1) * t))


def interpolate_hsb(xyz1: ColorTuple, xyz2: ColorTuple, t: float) -> ColorTuple:
    (h1, s1, b1) = xyz1
    (h2, s2, b2) = xyz2
    return (h1 + (h2 - h1) * t, s1 + (s2 - s1) * t, b1 + (b2 - b1) * t)


def rgb_distance(xyz1, xyz2):
    (r1, g1, b1) = xyz1
    (r2, g2, b2) = xyz2
    return int(abs(r1 - r2) + abs(g1 - g2) + abs(b1 - b2))



cdef extern from "vxl_c.cpp":
    enum:
        MAP_X
        MAP_Y
        MAP_Z
        DEFAULT_COLOR
    struct MapData:
        pass
    struct MapGenerator:
        pass
    MapGenerator * create_map_generator(MapData * original)
    void delete_map_generator(MapGenerator * generator)
    object get_generator_data(MapGenerator * generator, int columns)
    MapData * load_vxl(unsigned char * v)
    MapData * copy_map(MapData * map)
    void delete_vxl(MapData * map)
    object save_vxl(MapData * map)
    int check_node(int x, int y, int z, MapData * map, int destroy)
    bint get_solid(int x, int y, int z, MapData * map)
    int get_color(int x, int y, int z, MapData * map)
    void set_point(int x, int y, int z, MapData * map, bint solid, int color)
    void set_column_solid(int x, int y, int start_z, int end_z,
        MapData * map, bint solid)
    void set_column_color(int x, int y, int start_z, int end_z,
        MapData * map, int color)
    int get_random_point(int x1, int y1, int x2, int y2, MapData * map,
        float random_1, float random_2, int * x, int * y)
    bint is_valid_position(int x, int y, int z)
    void update_shadows(MapData * map)

cdef class VXLData:
    cdef MapData * map

    cpdef get_solid(self, int x, int y, int z)
    cpdef get_color(self, int x, int y, int z)
    cpdef tuple get_random_point(self, int x1, int y1, int x2, int y2)
    cpdef int get_z(self, int x, int y, int start = ?)
    cpdef int get_height(self, int x, int y)
    cpdef tuple get_safe_coords(self, int x, int y, int z)
    cpdef bint has_neighbors(self, int x, int y, int z)
    cpdef bint is_surface(self, int x, int y, int z)
    cpdef list get_neighbors(self, int x, int y, int z)
    cpdef int check_node(self, int x, int y, int z, bint destroy = ?)
    cpdef bint build_point(self, int x, int y, int z, tuple color)
    cpdef bint set_column_fast(self, int x, int y, int start_z,
        int end_z, int end_color_z, int color)
    cpdef update_shadows(self)



# READ THE INSTRUCTIONS BELOW BEFORE YOU ASK QUESTIONS

# Arena game mode written by Yourself A game of team survival. The last team
# standing scores a point.

# A map that uses arena needs to be modified to have a starting area for each
# team. A starting area is enclosed and has a gate on it. Each block of a gate
# must have the EXACT same color to work properly. Between each rounds, the
# gate is rebuilt. The gates are destroyed simultaneously at the start of each
# round, releasing the players onto the map. Players are free to switch weapons
# between rounds.

# Spawn locations and gate locations MUST be present in the map metadata (map
# txt file) for arena to work properly.

# The spawn location/s for the green team are set by using the data from the
# 'arena_green_spawns' tuple in the extensions dictionary. Likewise, the blue
# spawn/s is set with the 'arena_blue_spawns' key. 'arena_green_spawns' and
# 'arena_blue_spawns' are tuples which contain tuples of spawn coordinates.
# Spawn locations are chosen randomly.

# NOTE THAT THE SCRIPT RETAINS BACKWARDS COMPATIBILITY with the old
# 'arena_green_spawn' and 'arena_blue_spawn'

# The 'arena_max_spawn_distance' can be used to set MAX_SPAWN_DISTANCE on a map
# by map basis. See the comment by MAX_SPAWN_DISTANCE for more information

# The locations of gates is also determined in the map metadata. 'arena_gates'
# is a tuple of coordinates in the extension dictionary. Each gate needs only
# one block to be specified (since each gate is made of a uniform color)

# Sample extensions dictionary of an arena map with two gates: In this example
# there is one spawn location for blue and two spawn locations for green.
# extensions = { 'arena': True, 'arena_blue_spawns' : ((128, 256, 60),),
# 'arena_green_spawns' : ((384, 256, 60), (123, 423, 51)), 'arena_gates':
# ((192, 236, 59), (320, 245, 60)) }

import random
import math
from pyspades.contained import BlockAction, SetColor, BlockLine
from pyspades import world
from pyspades.constants import DESTROY_BLOCK, TEAM_CHANGE_KILL, CTF_MODE
from twisted.internet import reactor
from piqueserver.commands import command, admin

# If ALWAYS_ENABLED is False, then the 'arena' key must be set to True in
# the 'extensions' dictionary in the map metadata
ALWAYS_ENABLED = True

# How long should be spent between rounds in arena (seconds)
SPAWN_ZONE_TIME = 15.0

# How many seconds a team color should be shown after they win a round
# Set to 0 to disable this feature.
TEAM_COLOR_TIME = 4.0

# Maximum duration that a round can last. Time is in seconds. Set to 0 to
# disable the time limit
MAX_ROUND_TIME = 180

MAP_CHANGE_DELAY = 25.0

# Coordinates to hide the tent and the intel
HIDE_COORD = (0, 0, 63)

# Max distance a player can be from a spawn while the players are held within
# the gates. If they get outside this they are teleported to a spawn.
# Used to teleport players who glitch through the map back into the spawns.
MAX_SPAWN_DISTANCE = 15.0

BUILDING_ENABLED = False

if MAX_ROUND_TIME >= 60:
    MAX_ROUND_TIME_TEXT = '%.2f minutes' % (float(MAX_ROUND_TIME) / 60.0)
else:
    MAX_ROUND_TIME_TEXT = str(MAX_ROUND_TIME) + ' seconds'


@command(admin_only=True)
def coord(connection):
    connection.get_coord = True
    return 'Spade a block to get its coordinate.'


def make_color(r, g, b, a=255):
    r = int(r)
    g = int(g)
    b = int(b)
    a = float(a)
    return b | (g << 8) | (r << 16) | (int((a / 255.0) * 128.0) << 24)

# Algorithm for minimizing the number of blocks sent for the gates using
# a block line. Probably won't find the optimal solution for shapes that are not
# rectangular prisms but it's better than nothing.
# d = changing indice
# c1 = first constant indice
# c2 = second constant indice


def partition(points, d, c1, c2):
    row = {}
    row_list = []
    for point in points:
        pc1 = point[c1]
        pc2 = point[c2]
        if pc1 not in row:
            row[pc1] = {}
        dic1 = row[pc1]
        if pc2 not in dic1:
            dic1[pc2] = []
            row_list.append(dic1[pc2])
        dic2 = dic1[pc2]
        dic2.append(point)

    row_list_sorted = [sorted(div, key=lambda k: k[d]) for div in row_list]

    # row_list_sorted is a list containing lists of points that all have the same
    # point[c1] and point[c2] values and are sorted in increasing order
    # according to point[d]
    start_block = None
    final_blocks = []
    for block_list in row_list_sorted:
        counter = 0
        for i, block in enumerate(block_list):
            counter += 1
            if start_block is None:
                start_block = block
            if i + 1 == len(block_list):
                next_block = None
            else:
                next_block = block_list[i + 1]
            # Current AoS version seems to have an upper limit of 65 blocks for
            # a block line
            if counter == 65 or next_block is None or block[d] + \
                    1 != next_block[d]:
                final_blocks.append([start_block, block])
                start_block = None
                counter = 0
    return final_blocks


def minimize_block_line(points):
    x = partition(points, 0, 1, 2)
    y = partition(points, 1, 0, 2)
    z = partition(points, 2, 0, 1)
    xlen = len(x)
    ylen = len(y)
    zlen = len(z)
    if xlen <= ylen and xlen <= zlen:
        return x
    if ylen <= xlen and ylen <= zlen:
        return y
    if zlen <= xlen and zlen <= ylen:
        return z
    return x


def get_team_alive_count(team):
    count = 0
    for player in team.get_players():
        if player.world_object is not None:
            if not player.world_object.dead:
                count += 1
    return count


def get_team_dead(team):
    for player in team.get_players():
        if player.world_object is not None:
            if not player.world_object.dead:
                return False
    return True


class ArenaException(Exception):
    pass


class Gate:

    def __init__(self, x, y, z, protocol_obj):
        self.support_blocks = []
        self.blocks = []
        self.protocol_obj = protocol_obj
        solid, self.color = self.protocol_obj.map.get_point(x, y, z)
        if not solid:
            raise ArenaException(
                'The gate coordinate (%i, %i, %i) is not solid.' % (x, y, z))
        self.record_gate(x, y, z)
        self.blocks = minimize_block_line(self.blocks)

    def build_gate(self):
        map_ = self.protocol_obj.map
        block_line = BlockLine()
        set_color = SetColor()
        set_color.value = make_color(*self.color)
        set_color.player_id = block_line.player_id = 32
        self.protocol_obj.broadcast_contained(set_color, save=True)
        for block_line_ in self.blocks:
            start_block, end_block = block_line_
            points = world.cube_line(*(start_block + end_block))
            if not points:
                continue
            for point in points:
                x, y, z = point
                if not map_.get_solid(x, y, z):
                    map_.set_point(x, y, z, self.color)
            block_line.x1, block_line.y1, block_line.z1 = start_block
            block_line.x2, block_line.y2, block_line.z2 = end_block
            self.protocol_obj.broadcast_contained(block_line, save=True)

    def destroy_gate(self):
        map_ = self.protocol_obj.map
        block_action = BlockAction()
        block_action.player_id = 32
        block_action.value = DESTROY_BLOCK
        for block in self.support_blocks:  # optimize wire traffic
            if map_.get_solid(*block):
                map_.remove_point(*block)
                block_action.x, block_action.y, block_action.z = block
                self.protocol_obj.broadcast_contained(block_action, save=True)
        for block_line_ in self.blocks:  # avoid desyncs
            start_block, end_block = block_line_
            points = world.cube_line(*(start_block + end_block))
            if not points:
                continue
            for point in points:
                x, y, z = point
                if map_.get_solid(x, y, z):
                    map_.remove_point(x, y, z)

    def record_gate(self, x, y, z):
        if x < 0 or x > 511 or y < 0 or x > 511 or z < 0 or z > 63:
            return False
        solid, color = self.protocol_obj.map.get_point(x, y, z)
        if solid:
            coordinate = (x, y, z)
            if color[0] != self.color[0] or color[1] != self.color[1] or color[2] != self.color[2]:
                return True
            for block in self.blocks:
                if coordinate == block:
                    return False
            self.blocks.append(coordinate)
            returns = (self.record_gate(x + 1, y, z),
                       self.record_gate(x - 1, y, z),
                       self.record_gate(x, y + 1, z),
                       self.record_gate(x, y - 1, z),
                       self.record_gate(x, y, z + 1),
                       self.record_gate(x, y, z - 1))
            if True in returns:
                self.support_blocks.append(coordinate)
        return False


def apply_script(protocol, connection, config):
    class ArenaConnection(connection):
        get_coord = False

        def on_block_destroy(self, x, y, z, mode):
            returned = connection.on_block_destroy(self, x, y, z, mode)
            if self.get_coord:
                self.get_coord = False
                self.send_chat('Coordinate: %i, %i, %i' % (x, y, z))
                return False
            return returned

        def on_disconnect(self):
            if self.protocol.arena_running:
                if self.world_object is not None and not self.world_object.dead:
                    self.world_object.dead = True
                    self.protocol.check_round_end()
            return connection.on_disconnect(self)

        def on_kill(self, killer, kill_type, grenade):
            if self.protocol.arena_running and kill_type != TEAM_CHANGE_KILL:
                if self.world_object is not None and not self.world_object.dead:
                    self.world_object.dead = True
                    self.protocol.check_round_end(killer)
            return connection.on_kill(self, killer, kill_type, grenade)

        def on_team_join(self, team):
            returned = connection.on_team_join(self, team)
            if returned is False:
                return False
            if self.protocol.arena_running:
                if self.world_object is not None and not self.world_object.dead:
                    self.world_object.dead = True
                    self.protocol.check_round_end()
            return returned

        def on_position_update(self):
            if not self.protocol.arena_running:
                min_distance = None
                pos = self.world_object.position
                for spawn in self.team.arena_spawns:
                    xd = spawn[0] - pos.x
                    yd = spawn[1] - pos.y
                    zd = spawn[2] - pos.z
                    distance = math.sqrt(xd ** 2 + yd ** 2 + zd ** 2)
                    if min_distance is None or distance < min_distance:
                        min_distance = distance
                if min_distance > self.protocol.arena_max_spawn_distance:
                    self.set_location(random.choice(self.team.arena_spawns))
                    self.refill()
            return connection.on_position_update(self)

        def get_respawn_time(self):
            if self.protocol.arena_enabled:
                if self.protocol.arena_running:
                    return -1
                else:
                    return 0
            return connection.get_respawn_time(self)

        def respawn(self):
            if self.protocol.arena_running:
                return False
            return connection.respawn(self)

        def on_spawn(self, pos):
            returned = connection.on_spawn(self, pos)
            if self.protocol.arena_running:
                self.kill()
            return returned

        def on_spawn_location(self, pos):
            if self.protocol.arena_enabled:
                return random.choice(self.team.arena_spawns)
            return connection.on_spawn_location(self, pos)

        def on_flag_take(self):
            if self.protocol.arena_take_flag:
                self.protocol.arena_take_flag = False
                return connection.on_flag_take(self)
            return False

        def on_refill(self):
            returned = connection.on_refill(self)
            if self.protocol.arena_running:
                return False
            return returned

    class ArenaProtocol(protocol):
        game_mode = CTF_MODE
        old_respawn_time = None
        old_building = None
        old_killing = None
        arena_enabled = False
        arena_running = False
        arena_counting_down = False
        arena_take_flag = False
        arena_countdown_timers = None
        arena_limit_timer = None
        arena_old_fog_color = None
        arena_max_spawn_distance = MAX_SPAWN_DISTANCE

        def check_round_end(self, killer=None, message=True):
            if not self.arena_running:
                return
            for team in (self.green_team, self.blue_team):
                if get_team_dead(team):
                    self.arena_win(team.other, killer)
                    return
            if message:
                self.arena_remaining_message()

        def arena_time_limit(self):
            self.arena_limit_timer = None
            green_team = self.green_team
            blue_team = self.blue_team
            green_count = get_team_alive_count(green_team)
            blue_count = get_team_alive_count(blue_team)
            if green_count > blue_count:
                self.arena_win(green_team)
            elif green_count < blue_count:
                self.arena_win(blue_team)
            else:
                self.send_chat('Round ends in a tie.')
            self.begin_arena_countdown()

        def arena_win(self, team, killer=None):
            if not self.arena_running:
                return
            if self.arena_old_fog_color is None and TEAM_COLOR_TIME > 0:
                self.arena_old_fog_color = self.fog_color
                self.set_fog_color(team.color)
                reactor.callLater(TEAM_COLOR_TIME, self.arena_reset_fog_color)
            if killer is None or killer.team is not team:
                for player in team.get_players():
                    if player.world_object is not None and not player.world_object.dead:
                        killer = player
                        break
            if killer is not None:
                self.arena_take_flag = True
                killer.take_flag()
                killer.capture_flag()
            self.send_chat(team.name + ' team wins the round!')
            self.begin_arena_countdown()

        def arena_reset_fog_color(self):
            if self.arena_old_fog_color is not None:
                # Shitty fix for disco on game end
                self.old_fog_color = self.arena_old_fog_color
                self.set_fog_color(self.arena_old_fog_color)
                self.arena_old_fog_color = None

        def arena_remaining_message(self):
            if not self.arena_running:
                return
            green_team = self.green_team
            blue_team = self.blue_team
            for team in (self.green_team, self.blue_team):
                num = get_team_alive_count(team)
                team.arena_message = '%i player' % num
                if num != 1:
                    team.arena_message += 's'
                team.arena_message += ' on ' + team.name
            self.send_chat('%s and %s remain.' %
                           (green_team.arena_message, blue_team.arena_message))

        def on_map_change(self, map_):
            extensions = self.map_info.extensions
            if ALWAYS_ENABLED:
                self.arena_enabled = True
            else:
                if 'arena' in extensions:
                    self.arena_enabled = extensions['arena']
                else:
                    self.arena_enabled = False
            self.arena_max_spawn_distance = MAX_SPAWN_DISTANCE
            if self.arena_enabled:
                self.old_respawn_time = self.respawn_time
                self.respawn_time = 0
                self.old_building = self.building
                self.old_killing = self.killing
                self.gates = []
                if 'arena_gates' in extensions:
                    for gate in extensions['arena_gates']:
                        self.gates.append(Gate(*gate, protocol_obj=self))
                if 'arena_green_spawns' in extensions:
                    self.green_team.arena_spawns = extensions[
                        'arena_green_spawns']
                elif 'arena_green_spawn' in extensions:
                    self.green_team.arena_spawns = (
                        extensions['arena_green_spawn'],)
                else:
                    raise ArenaException(
                        'No arena_green_spawns given in map metadata.')
                if 'arena_blue_spawns' in extensions:
                    self.blue_team.arena_spawns = extensions[
                        'arena_blue_spawns']
                elif 'arena_blue_spawn' in extensions:
                    self.blue_team.arena_spawns = (
                        extensions['arena_blue_spawn'],)
                else:
                    raise ArenaException(
                        'No arena_blue_spawns given in map metadata.')
                if 'arena_max_spawn_distance' in extensions:
                    self.arena_max_spawn_distance = extensions[
                        'arena_max_spawn_distance']
                self.delay_arena_countdown(MAP_CHANGE_DELAY)
                self.begin_arena_countdown()
            else:
                # Cleanup after a map change
                if self.old_respawn_time is not None:
                    self.respawn_time = self.old_respawn_time
                if self.old_building is not None:
                    self.building = self.old_building
                if self.old_killing is not None:
                    self.killing = self.old_killing
                self.arena_enabled = False
                self.arena_running = False
                self.arena_counting_down = False
                self.arena_limit_timer = None
                self.arena_old_fog_color = None
                self.old_respawn_time = None
                self.old_building = None
                self.old_killing = None
            return protocol.on_map_change(self, map)

        def build_gates(self):
            for gate in self.gates:
                gate.build_gate()

        def destroy_gates(self):
            for gate in self.gates:
                gate.destroy_gate()

        def arena_spawn(self):
            for player in list(self.players.values()):
                if player.team.spectator:
                    continue
                if player.world_object is not None and player.world_object.dead:
                    player.spawn(random.choice(player.team.arena_spawns))
                else:
                    player.set_location(
                        random.choice(player.team.arena_spawns))
                    player.refill()

        def refill_all(self):
            for player in self.players.values():
                if player.team.spectator:
                    continue
                player.refill()

        def begin_arena_countdown(self):
            if self.arena_limit_timer is not None:
                if self.arena_limit_timer.cancelled == 0 and self.arena_limit_timer.called == 0:
                    self.arena_limit_timer.cancel()
                    self.arena_limit_timer = None
            if self.arena_counting_down:
                return
            self.arena_running = False
            self.arena_limit_timer = None
            self.arena_counting_down = True
            self.killing = False
            self.building = False
            self.build_gates()
            self.arena_spawn()
            self.send_chat('The round will begin in %i seconds.' %
                           SPAWN_ZONE_TIME)
            self.arena_countdown_timers = [
                reactor.callLater(SPAWN_ZONE_TIME, self.begin_arena)]
            for time in range(1, 6):
                self.arena_countdown_timers.append(reactor.callLater(
                    SPAWN_ZONE_TIME - time, self.send_chat, str(time)))

        def delay_arena_countdown(self, amount):
            if self.arena_counting_down:
                for timer in self.arena_countdown_timers:
                    if timer.cancelled == 0 and timer.called == 0:
                        timer.delay(amount)

        def begin_arena(self):
            self.arena_counting_down = False
            for team in (self.green_team, self.blue_team):
                if team.count() == 0:
                    self.send_chat(
                        'Not enough players on the %s team to begin.' %
                        team.name)
                    self.begin_arena_countdown()
                    return
            self.arena_running = True
            self.killing = True
            self.building = BUILDING_ENABLED
            self.refill_all()
            self.destroy_gates()
            self.send_chat('Go!')
            if MAX_ROUND_TIME > 0:
                self.send_chat(
                    'There is a time limit of %s for this round.' %
                    MAX_ROUND_TIME_TEXT)
                self.arena_limit_timer = reactor.callLater(
                    MAX_ROUND_TIME, self.arena_time_limit)

        def on_base_spawn(self, x, y, z, base, entity_id):
            if not self.arena_enabled:
                return protocol.on_base_spawn(self, x, y, z, base, entity_id)
            return HIDE_COORD

        def on_flag_spawn(self, x, y, z, flag, entity_id):
            if not self.arena_enabled:
                return protocol.on_base_spawn(self, x, y, z, flag, entity_id)
            return HIDE_COORD

    return ArenaProtocol, ArenaConnection



"""
Kicks jerks for 'PRESS ALT-F4 FOR AIRSTRIKES' and so on.

Options
^^^^^^^

.. code-block:: guess

   [antijerk]
   ban_duration = "15min"

.. codeauthor:: ?
"""

import re
from piqueserver.config import config, cast_duration

chat_pattern = re.compile(".*(airstrike).*(esc|escape|alt-f4|alt f4)",
                          re.IGNORECASE)
chat_pattern_2 = re.compile(".*(esc|escape|alt-f4|alt f4).*(airstrike)",
                            re.IGNORECASE)
admin_pattern = re.compile(".*(admin)",
                           re.IGNORECASE)
antijerk_config = config.section("antijerk")
ban_duration = antijerk_config.option("ban_duration", default="15min", cast=cast_duration)


def antijerk_match(player, msg):
    if not (player.user_types.trusted or player.admin):
        return chat_pattern.match(msg) or chat_pattern_2.match(msg) or\
            admin_pattern.match(player.name)


def apply_script(protocol, connection, config):
    def jerk_kick(connection):
        connection.ban('Autoban: anti-jerk', ban_duration.get())

    class AntiJerkConnection(connection):

        def on_chat(self, value, global_message):
            if antijerk_match(self, value):
                jerk_kick(self)
            else:
                return connection.on_chat(self, value, global_message)

        def on_login(self, name):
            if admin_pattern.match(name):
                self.send_chat('Anti-jerk: Please remove "Admin" from your ' +
                               'name: TALKING WILL AUTOBAN YOU!')
            return connection.on_login(self, name)

    return protocol, AntiJerkConnection



"""
Airstrikes. Boom!

Commands
^^^^^^^^

* ``/airstrike or /a`` allows a player to call an airstike
* ``/givestrike <player>`` gives a player the ability to call an airstike immediately *admin only*

.. codeauthor:: hompy
"""

from math import ceil, sin, cos
from random import uniform, vonmisesvariate
from twisted.internet import reactor
from pyspades.contained import GrenadePacket
from pyspades.common import to_coordinates, Vertex3
from pyspades.world import Grenade
from pyspades.constants import UPDATE_FREQUENCY, WEAPON_TOOL
from piqueserver.commands import (
    command, admin, get_player, player_only, target_player)

STREAK_REQUIREMENT = 8


REFILL_ON_GRANT = True  # heals when unlocking the airstrike
REFILL_ON_AIRSTRIKE = False  # heals when calling in the airstrike


S_READY = "Airstrike support ready! SCOPE and HOLD <V> (sneak) to launch"
S_FAILED = 'You need to aim your airstrike somewhere, not at the sky!'
S_NO_STREAK = 'Every {streak} kills in a row you unlock an ' \
    'airstrike. {remaining} kills to go!'
S_COOLDOWN = '{seconds} seconds before your team can launch another airstrike'
S_ALLIED = 'Ally {player} called in an airstrike on location {coords}'
S_ENEMY = '[WARNING] Enemy air support heading to {coords}!'
S_MOVED = 'You must hold still while targeting the airstrike'
S_STAND = 'You must be standing up to target the airstrike'


TEAM_COOLDOWN = 30.0  # seconds until another airstrike can be called
ZOOMV_TIME = 0.1  # seconds holding V until airstrike is called
ZOOMV_RAY_LENGTH = 192.0
ARRIVAL_DELAY = 2  # seconds from airstrike notice to arrival


@command('airstrike', 'a')
@player_only
def airstrike(connection, *args):
    player = connection

    if player.airstrike:
        return S_READY

    kills_left = STREAK_REQUIREMENT - player.airstrike_streak
    return S_NO_STREAK.format(streak=STREAK_REQUIREMENT,
                              remaining=kills_left)

# debug


@command('givestrike', admin_only=True)
@target_player
def give_strike(connection, player):
    player.airstrike = True
    player.send_chat(S_READY)


def bellrand(a, b):
    return (uniform(a, b) + uniform(a, b) + uniform(a, b)) / 3.0


class Nag:
    call = None

    def __init__(self, secs, f, *args, **kw):
        self.seconds = secs
        self.f = f
        self.args = args
        self.kw = kw

    def start_or_reset(self):
        if self.call and self.call.active():
            # In Twisted==18.9.0, reset() is broken when using
            # AsyncIOReactor
            # self.call.reset(ZOOMV_TIME)
            self.call.cancel()
            self.call = reactor.callLater(
                ZOOMV_TIME, self.f, *self.args, **self.kw)
        else:
            self.call = reactor.callLater(
                self.seconds, self.f, *self.args, **self.kw)

    def stop(self):
        if self.call and self.call.active():
            self.call.cancel()
        self.call = None

    def active(self):
        return self.call and self.call.active()

    def __del__(self):
        self.stop()


def apply_script(protocol, connection, config):
    class AirstrikeConnection(connection):
        airstrike = False
        airstrike_grenade_calls = None
        airstrike_streak = 0
        airstrike_nag = None
        zoomv = None
        zoomv_nag = None
        last_zoomv_message = None

        def start_airstrike(self, x, y, z):
            coords = to_coordinates(x, y)
            message = S_ALLIED.format(player=self.name, coords=coords)
            self.protocol.send_chat(message, global_message=False,
                                    team=self.team)
            message = S_ENEMY.format(coords=coords)
            self.protocol.send_chat(message, global_message=False,
                                    team=self.team.other)
            self.team.last_airstrike = reactor.seconds()

            reactor.callLater(ARRIVAL_DELAY, self.do_airstrike, x, y, z)

        def do_airstrike(self, x, y, z):
            if self.name is None:
                return
            direction = 1 if self.team.id == 0 else -1
            # forward randomization range
            jitter = (3.0 * direction, 4.0 * direction)
            spread = 0.6  # symmetric Y up and down
            radius = 30.0  # maximum reach from ground zero
            estimate_travel = 32 * z / 64  # to offset spawn coordinates
            x = max(0, min(511, x - estimate_travel * direction))
            grenade_z = 1

            self.airstrike_grenade_calls = []
            for i in range(10):
                angle = vonmisesvariate(0.0, 0.0)  # fancy for uniform(0, 2*pi)
                distance = bellrand(-radius, radius)
                grenade_x = x + cos(angle) * distance
                grenade_y = y + sin(angle) * distance
                for j in range(5):
                    grenade_x += uniform(*jitter)
                    grenade_y += uniform(-spread, spread)
                    delay = i * 0.85 + j * 0.11
                    call = reactor.callLater(
                        delay,
                        self.create_airstrike_grenade,
                        grenade_x,
                        grenade_y,
                        grenade_z)
                    self.airstrike_grenade_calls.append(call)

        def create_airstrike_grenade(self, x, y, z):
            if self.name is None:
                return
            direction = 1 if self.team.id == 0 else -1
            z_speed = 3.5  # used to be 0.5

            position = Vertex3(x, y, z)
            velocity = Vertex3(direction, 0.0, z_speed)
            grenade = self.protocol.world.create_object(
                Grenade, 0.0, position, None, velocity, self.airstrike_exploded)
            grenade.name = 'airstrike'

            collision = grenade.get_next_collision(UPDATE_FREQUENCY)
            if not collision:
                return
            eta, x, y, z = collision
            grenade.fuse = eta
            grenade_packet = GrenadePacket()
            grenade_packet.value = grenade.fuse
            grenade_packet.player_id = self.player_id
            grenade_packet.position = position.get()
            grenade_packet.velocity = velocity.get()
            self.protocol.broadcast_contained(grenade_packet)

        def airstrike_exploded(self, grenade):
            grenade.velocity.normalize()
            penetration = 3  # go forward some extra distance until we collide
            while penetration:
                penetration -= 1
                grenade.position += grenade.velocity
                solid = self.protocol.map.get_solid(*grenade.position.get())
                if solid or solid is None:
                    break
            self.grenade_exploded(grenade)

        def end_airstrike(self):
            if self.airstrike_grenade_calls:
                for grenade_call in self.airstrike_grenade_calls:
                    if grenade_call and grenade_call.active():
                        grenade_call.cancel()
            self.airstrike_grenade_calls = None

        def start_zoomv(self):
            now = reactor.seconds()
            last_strike = getattr(self.team, 'last_airstrike', None)
            if last_strike is not None and now - last_strike < TEAM_COOLDOWN:
                remaining = ceil(TEAM_COOLDOWN - (now - last_strike))
                message = S_COOLDOWN.format(seconds=int(remaining))
                self.send_zoomv_chat(message)
                return
            location = self.world_object.cast_ray(ZOOMV_RAY_LENGTH)
            if not location:
                self.send_zoomv_chat(S_FAILED)
                return
            self.zoomv.start_or_reset()

        def end_zoomv(self):
            location = self.world_object.cast_ray(ZOOMV_RAY_LENGTH)
            if not location:
                self.send_chat(S_FAILED)
                return
            self.airstrike = False
            self.airstrike_streak = 0
            self.start_airstrike(*location)
            if REFILL_ON_AIRSTRIKE:
                self.refill()

        def send_zoomv_chat(self, message):
            last_message = self.last_zoomv_message
            if last_message is None or reactor.seconds() - last_message >= 1.0:
                self.send_chat(message)
                self.last_zoomv_message = reactor.seconds()

        def on_team_changed(self, old_team):
            self.end_airstrike()
            connection.on_team_changed(self, old_team)

        def on_login(self, player_name):
            self.airstrike_nag = Nag(4.0, self.send_chat, S_READY)
            self.zoomv = Nag(1.0, self.end_zoomv)
            self.zoomv_nag = Nag(2.0, self.send_zoomv_chat, S_STAND)
            connection.on_login(self, player_name)

        def on_reset(self):
            self.airstrike = False
            self.airstrike_streak = 0
            self.airstrike_nag = None
            self.zoomv = None
            self.zoomv_nag = None
            self.end_airstrike()
            connection.on_reset(self)

        def on_spawn(self, location):
            if self.airstrike:
                self.airstrike_nag.start_or_reset()
            connection.on_spawn(self, location)

        def on_kill(self, killer, kill_type, grenade):
            self.airstrike_streak = 0
            self.airstrike_nag.stop()
            self.zoomv.stop()
            if killer is not None and self.team is not killer.team:
                # no self-kills and no teamkills
                if grenade is None or grenade.name == 'grenade':
                    # only regular grenades count for streak
                    if not killer.airstrike:
                        # don't increase streak if we have a strike on-hold
                        killer.airstrike_streak += 1
                        if killer.airstrike_streak == STREAK_REQUIREMENT:
                            killer.airstrike = True
                            killer.send_chat(S_READY)
                            if REFILL_ON_GRANT:
                                killer.refill()
            connection.on_kill(self, killer, kill_type, grenade)

        def on_shoot_set(self, fire):
            if self.zoomv.active() and fire:
                self.zoomv.start_or_reset()
                self.send_zoomv_chat(S_MOVED)
            connection.on_shoot_set(self, fire)

        def on_walk_update(self, up, down, left, right):
            if self.zoomv.active() and (up or down or left or right):
                self.zoomv.start_or_reset()
                self.send_zoomv_chat(S_MOVED)
            return connection.on_walk_update(self, up, down, left, right)

        def on_secondary_fire_set(self, secondary):
            if self.airstrike and self.tool == WEAPON_TOOL:
                if secondary:
                    if self.world_object.crouch:
                        self.zoomv_nag.start_or_reset()
                else:
                    self.zoomv.stop()
                    self.zoomv_nag.stop()
            connection.on_secondary_fire_set(self, secondary)

        def on_animation_update(self, jump, crouch, sneak, sprint):
            obj = self.world_object
            secondary = obj.secondary_fire
            if self.airstrike and self.tool == WEAPON_TOOL and secondary:
                if sneak and not self.zoomv.active():
                    if obj.up or obj.down or obj.left or obj.right:
                        self.send_zoomv_chat(S_MOVED)
                        return
                    self.start_zoomv()
                elif not sneak and self.zoomv.active():
                    if crouch:
                        # crouching was the reason sneak got turned off
                        self.send_zoomv_chat(S_STAND)
                    self.zoomv.stop()
                elif crouch:
                    self.zoomv_nag.start_or_reset()
                elif not crouch:
                    self.zoomv_nag.stop()
            return connection.on_animation_update(self, jump, crouch, sneak,
                                                  sprint)

    return protocol, AirstrikeConnection



"""
Prevents taking intel through walls

Version 2(2017.12.25)

.. codeauthor:: kmsi<kmsiapps@gmail.com>
"""


def apply_script(protocol, connection, config):
    class noIntelWallsConnection(connection):

        def on_flag_take(self):
            flag = self.team.other.flag
            if not self.world_object.can_see(flag.x, flag.y, flag.z-1):
                return False
            else:
                return connection.on_flag_take(self)
    return protocol, noIntelWallsConnection
